
local addon_name, addon = ...

local _G = _G
_G.PetLeash = LibStub("AceAddon-3.0"):NewAddon(addon, addon_name, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("PetLeash")
local LibPetJournal = LibStub("LibPetJournal-2.0")

local ipairs, pairs, tonumber, tostring = _G.ipairs, _G.pairs, _G.tonumber, _G.tostring
local assert, tinsert, wipe = _G.assert, _G.tinsert, _G.wipe

--
-- Binding globals
--

_G.BINDING_HEADER_PETLEASH = addon_name
_G.BINDING_NAME_PETLEASH_SUMMON = L["Summon Another Pet"]
_G.BINDING_NAME_PETLEASH_DESUMMON = L["Desummon Pet"]
_G.BINDING_NAME_PETLEASH_TOGGLE = L["Toggle Non-Combat Pet"]
_G.BINDING_NAME_PETLEASH_CONFIG = L["Open Configuration"]

--
-- Default DB
--
local defaults = {
    profile = {
        enable = true,
        enableInCombat = false,
        enableInBattleground = true,
        dismissInBattleground = false,
        disableOutsideCities = false,
        dismissWhenStealthed = true,
        dismissWhileFlying = false,
        disableForQuestItems = true,
        overrideSetLoadout = false,
        waitTimer = 3,
        verbose = false,
        weightedPets = false,
        selectedQuestItems = {}, -- "item:xx" = true, disable.  if false, use check
        autoSwitch = {
            timer = false,
            timerValue = 30*60,
            citiesOnly = false,
            onZone = false,
        },
        default_ignore = false,
        useFavorites = false,
        ignore_pets = {         -- [petid] = true, don't use.  false, use.  nil, no opinion (see default_ignore)
        },
        weights = {},           -- [spellid] = num (if nil default is 1)
    }
}

local default_ignore_override = {
        [15429] = true,     -- Disgusing Oozling (Combat Effect)
        [49590] = true,     -- Guild Page, Horde (Long cooldown)
        [49588] = true,     -- Guild Herald, Horde (Long cooldown)
        [49586] = true,     -- Guild Page, Alliance (Long cooldown)
        [49587] = true,     -- Guild Herald, Alliance (Long cooldown)
}


-- config

local function config_toggle_get(info) return addon.db.profile[info[#info]] end
local function config_toggle_set(info,v) addon.db.profile[info[#info]] = v addon:UpdatePetList() addon.Ready:UpdateChecks() end

local options = {
    name = "PetLeash",
    handler = PetLeash,
    type = 'group',
    args = {
        main = {
            name = GENERAL,
            type = 'group',
            childGroups = "tab",
            args = {
                general = {
                    name = GENERAL,
                    type = "group",
                    order = 10,
                    args = {
                        enable = {
                            type = "toggle",
                            name = ENABLE,
                            desc = L["Enable Auto-Summon"], 
                            order = 10,
                            width = "full",
                            get = function(info) return addon:IsEnabledSummoning() end,
                            set = function(info,v) addon:EnableSummoning(v) end,
                        },
                        enableInCombat = {
                            type = "toggle",
                            name = L["Enable In Combat"],
                            order = 11,
                            width = "double",
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        enableInBattleground = {
                            type = "toggle",
                            name = L["Enable In Battlegrounds/Arena"],
                            order = 12,
                            width = "double",
                            disabled = function() return addon.db.profile.dismissInBattleground end,
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        disableOutsideCities = {
                            type = "toggle",
                            name = L["Only Enable in Cities"],
                            order = 13,
                            width = "double",
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        dismissWhenStealthed = {
                            type = "toggle",
                            name = L["Dismiss When Stealthed or Invisible"],
                            order = 14,
                            width = "double",
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        dismissWhileFlying = {
                            type = "toggle",
                            name = L["Dismiss When Flying"],
                            order = 15,
                            width = "double",
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        dismissInBattleground = {
                            type = "toggle",
                            name = L["Dismiss In Battlegrounds/Arena"],
                            order = 16,
                            width = "double",
                            disabled = function() return addon.db.profile.enableInBattleground end,
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        overrideSetLoadout = {
                            type = "toggle",
                            name = L["Override Pet Battle Loadout"],
                            desc = L["Resummon pet after a pet has been summoned via setting the battle pet loadout."],
                            order = 30,
                            width = "double",
                            get = config_toggle_get,
                            set = config_toggle_set
                        },
                        waitTimerValue = {
                            type = "range",
                            name = L["Wait Time (Seconds)"],
                            desc = L["How long must pass before a player is considered idle enough to summon a pet."],
                            order = 40,
                            min = 1,
                            step = .5,
                            bigStep = 1,
                            max = 30,
                            get = function()
                                return addon.db.profile.waitTimer
                            end,
                            set = function(info,v)
                                addon.db.profile.waitTimer = v
                            end  
                        },
                        verbose = {
                            type = "toggle",
                            name = L["Verbose"],
                            order = 100,
                            width = "double",
                            get = config_toggle_get,
                            set = config_toggle_set
                        }
                    },
                },
                autoSwitch = {
                    name = L["Auto Switch Pet"],
                    type = "group",
                    order = 30,
                    args = {
                        autoSwitchTimerEnable = {
                            type = "toggle",
                            name = L["Enable Timed Auto Switch"],
                            width = "double",
                            order = 31,
                            get = function() return addon.db.profile.autoSwitch.timer end,
                            set = function(info,v)
                                addon.db.profile.autoSwitch.timer = v
                                if(v) then
                                    addon.SwitchTimer:Start()
                                end
                            end
                        },
                        autoSwitchTimerValue = {
                            type = "range",
                            name = L["Seconds between switch"],
                            order = 32,
                            min = 60,
                            step = 1,
                            bigStep = 60,
                            max = 3600,
                            disabled = function() return not addon.db.profile.autoSwitch.timer end,
                            get = function()
                                return addon.db.profile.autoSwitch.timerValue
                            end,
                            set = function(info,v)
                                addon.db.profile.autoSwitch.timerValue = v
                                addon.SwitchTimer:Start()
                            end
                        },
                        autoSwitchCitiesOnly = {
                            type = "toggle",
                            name = L["Only use Timed Auto Switch in cities"],
                            width = "double",
                            order = 33,
                            get = function() return addon.db.profile.autoSwitch.citiesOnly end,
                            set = function(info,v) addon.db.profile.autoSwitch.citiesOnly = v end,
                        },
                        autoSwitchOnZone = {
                            type = "toggle",
                            name = L["Auto Switch when changing maps"],
                            width = "double",
                            order = 40,
                            get = function() return addon.db.profile.autoSwitch.onZone end,
                            set = function(info,v) addon.db.profile.autoSwitch.onZone = v end,
                        }
                    }
                }
            }
        },
        pets = {
            type = "group",
            name = L["Enabled Pets"],
            order = 10,
            cmdHidden = true,
            args = {
                enableAll = {
                    type = "execute",
                    name = L["Enable All"],
                    order = 1,
                    func = function(info)
                        addon:_Config_PetToggle_SetAll(info, false)
                    end
                },
                disableAll = {
                    type = "execute",
                    name = L["Disable All"],
                    order = 2,
                    func = function(info)
                        addon:_Config_PetToggle_SetAll(info, true)
                    end
                },
                useWeightedPets = {
                    -- If we get more pickers, change this to type = "select"
                    type = "toggle",
                    name = L["Weighted Pets"],
                    order = 3,
                    disabled = function() return addon.db.profile.useFavorites end,
                    get = function(info)
                        return addon.db.profile.weightedPets
                    end,
                    set = function(info, v)
                        addon.db.profile.weightedPets = v
                        addon:UpdateConfigTables()
                    end
                },
                useFavorites = {
                    type = "toggle",
                    name = L["Use Favorites"],
                    order = 4,
                    disabled = function() return addon.db.profile.weightedPets end,
                    get = function(info)
                        return addon.db.profile.useFavorites
                    end,
                    set = function(info, v)
                        addon.db.profile.useFavorites = v
                        addon:UpdatePetList(true)
                    end 
                },
                default_ignore = {
                    type = "toggle",
                    name = DEFAULT,
                    order = 5,
                    get = function(info)
                        return not addon.db.profile.default_ignore
                    end,
                    set = function(info, v)
                        addon.db.profile.default_ignore = not v
                        addon:UpdatePetList(true)
                    end 
                },
                seperator = {
                    type = "header",
                    name = "",
                    order = 9,
                },
                pets = {
                    type = "group",
                    name = "",
                    order = 10,
                    args = {},
                    inline = true
                }
            }
        },
    },
    plugins = {}
}

local options_slashcmd = {
    name = "PetLeash Slash Command",
    handler = PetLeash,
    type = "group",
    order = -2,
    args = {
        config = {
            type = "execute",
            name = L["Open Configuration"],
            dialogHidden = true,
            order = 1,
            func = function(info) addon:OpenOptions() end
        },
        resummon = {
            type = "execute",
            name = L["Summon Another Pet"],
            desc = L["Desummon your current pet and summon another pet.  Enable summoning if needed."],
            order = 20,
            func = function(info) addon:ResummonPet(true, true) end
        },
        desummon = {
            type = "execute",
            name = L["Desummon Pet"],
            desc = L["Desummon your currently summoned pet.  Disable summoning."],
            order = 21,
            func = function(info) addon:DesummonPet(true) end
        },
        togglePet = {
            type = "execute",
            name = L["Toggle Non-Combat Pet"],
            order = 22,
            func = function(info) addon:TogglePet() end
        },
        enable = options.args.main.args.general.args.enable,
        enableInCombat = options.args.main.args.general.args.enableInCombat,
        dismissWhenStealthed = options.args.main.args.general.args.dismissWhenStealthed,
        disableForQuestItems = options.args.main.args.general.args.disableForQuestItems,
        debugstate = {
            type = "execute",
            name = "Debug State",
            hidden = true,
            func = function(info) addon:DumpDebugState() end
        }
    },
}

local AceConfigRegistry = LibStub("AceConfigRegistry-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

function addon:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("PetLeash2DB", defaults, true)
    self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChange")
    self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChange")
    self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChange")

    self.usable_pets = {}   -- {petid, petid, ...}

    self.options = options
    self.options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    self.options_slashcmd = options_slashcmd
    
    AceConfig:RegisterOptionsTable(self.name, options)
    AceConfig:RegisterOptionsTable(self.name .. "SlashCmd", options_slashcmd, {"petleash", "pl"})
 
    self.optionsFrame = AceConfigDialog:AddToBlizOptions(self.name, self.name, nil, "main")
    self.optionsFrame:HookScript("OnShow", function() self:SetupOptions() end)
    InterfaceOptionsFrame:HookScript("OnShow", function() self:SetupOptions() end)
    
    hooksecurefunc(C_PetJournal, "SetPetLoadOutInfo", function(...) self:OnSetPetLoadOut(...) end)
    
    LibPetJournal.RegisterCallback(self, "PetListUpdated")
    self:PetListUpdated()
end

function addon:SetupOptions()
    -- we delay setting up options until here so that plugins in seperate
    -- addons will load properly
    if self.optionsFrame._finished then return end
    self.optionsFrame._finished = true
    
    self.optionsFrame.Pets = AceConfigDialog:AddToBlizOptions(self.name, L["Enabled Pets"], self.name, "pets")
    for name, module in self:IterateModules() do
        local f = module["SetupOptions"]
        if f then
            f(module, function(appName, name) AceConfigDialog:AddToBlizOptions(appName, name, self.name) end)
        end
    end
    self.optionsFrame.Profile = AceConfigDialog:AddToBlizOptions(self.name, L["Profiles"], self.name, "profiles")
    self.optionsFrame.About = LibStub("LibAboutPanel").new(self.name, self.name)
end

function addon:IsEnabledSummoning()
    return self.db.profile.enable
end

function addon:EnableSummoning(v)
    local oldv = self.db.profile.enable

    if((not oldv) ~= (not v)) then
        self.db.profile.enable = v
        
        -- TODO: is there a better way to trigger config update?
        AceConfigRegistry:NotifyChange("PetLeash")
        
        self:SendMessage("PetLeash-EnableState", v)
    end
end

function addon:OpenOptions()
    self:SetupOptions()
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
end

function addon:DumpDebugState()
    if self:IsEnabledSummoning() then
        self:Printf("Enabled (usable=%d)",#self.usable_pets)
    else
        self:Printf("Disabled (usable=%d)",#self.usable_pets)
    end    
    
    for name, module in self:IterateModules() do
        local f = module["DumpDebugState"]
        if f then
            f(module)
        end
    end
end

local BATTLEGROUND_ARENA = {["pvp"] = 1, ["arena"] = 1}
local function InBattlegroundOrArena()
    local _,t = IsInInstance()
    return BATTLEGROUND_ARENA[t]
end
addon.InBattlegroundOrArena = InBattlegroundOrArena

local RAID_INSTANCE = {["party"] = 1, ["raid"] = 1}
local function InInstanceOrRaid()
    local _,t = IsInInstance()
    return RAID_INSTANCE[t]
end
addon.InInstanceOrRaid = InInstanceOrRaid

-- pet list handling

function addon:IsPetIgnored(petid)
    if self.db.profile.useFavorites then
        local _, _, _, _, _, _, petName = C_PetJournal.GetPetInfoByPetID(petid)
        return not C_PetJournal.PetIsFavorite(petid)
    end
    
    local ignore = self.db.profile.ignore_pets[petid]
    if ignore ~= nil then
        return ignore
    end
    
    local _, _, _, _, _, _, _, _, _, creatureid = C_PetJournal.GetPetInfoByPetID(petid)
    if default_ignore_override[creatureid] then
        return true
    end
    
    return self.db.profile.default_ignore
end

function addon:UpdatePetList(updateconfig)
    wipe(self.usable_pets)
    
    for i,petID in LibPetJournal:IteratePetIDs() do
        local isSummonable = C_PetJournal.PetIsSummonable(petID)
        if isSummonable and not self:IsPetIgnored(petID) then
            tinsert(self.usable_pets, petID)
        end
    end
    
    if updateconfig == nil or updateconfig then
        self:UpdateConfigTables(true)
    end
end

function addon:PetListUpdated()
    self:UpdatePetList()
end

function addon:OnProfileChange()
    self:UpdatePetList()
    self.Ready:UpdateChecks(self.db.profile)
end

local L_WeightValues = {"|cffff0000"..L["Never"].."|r",
                        "|cffff6600"..L["Hardly Ever"].."|r",
                        "|cffff9900"..L["Rarely"].."|r",
                        "|cffddff00"..L["Occasionally"].."|r",
                        "|cff99ff00"..L["Sometimes"].."|r",
                        "|cff00ff00"..L["Often"].."|r"}

function addon:UpdateConfigTables()
    local args = options.args.pets.args.pets.args
    local useWeighted = self.db.profile.weightedPets
    
    wipe(args)
    
    for i,petID in LibPetJournal:IteratePetIDs() do
        local _, customName, _, _, _, _, petName = C_PetJournal.GetPetInfoByPetID(petID)
        local usename = customName or petName
                
        if(not useWeighted) then
            args[tostring(petID)] = {
                type = "toggle",
                name = usename,
                order = 1,
                get = "Config_PetToggle_Get",
                set = "Config_PetToggle_Set",
                hidden = "Config_PetToggle_Hidden"
            }
        else
            args[tostring(petID)] = {
                type = "select",
                name = usename,
                order = 1,
                values = L_WeightValues,
                get = "Config_PetToggle_Weighted_Get",
                set = "Config_PetToggle_Weighted_Set",
                hidden = "Config_PetToggle_Hidden"
            }
        end
    end
    
    -- Config Tables changed!
    AceConfigRegistry:NotifyChange("PetLeash")
end

function addon:Config_PetToggle_Hidden()
    return addon.db.profile.useFavorites
end

function addon:Config_PetToggle_Set(info, v)
    if(v) then
        self.db.profile.ignore_pets[tonumber(info[#info])] = false
    else
        self.db.profile.ignore_pets[tonumber(info[#info])] = true
    end
    
    self:UpdatePetList(false)
end

function addon:Config_PetToggle_Get(info)
    return not self:IsPetIgnored(tonumber(info[#info]))
end

function addon:Config_PetToggle_Weighted_Get(info)
    local id = tonumber(info[#info])
    if not self:IsPetIgnored(id) then
        return math.floor((self.db.profile.weights[id] or 1)*5)+1
    end
    return 1
end

function addon:Config_PetToggle_Weighted_Set(info, v)
    local id = tonumber(info[#info])
    if(v == 1) then
        self.db.profile.ignore_pets[id] = true
    else
        self.db.profile.ignore_pets[id] = false
        self.db.profile.weights[id] = (v-1)/5
    end
    self:UpdatePetList(false)
end

function addon:_Config_PetToggle_SetAll(info, v)
    for key in pairs(info.options.args.pets.args.pets.args) do
        self.db.profile.ignore_pets[tonumber(key)] = v
    end
    self:UpdatePetList(false)
end

--
--
--

function addon:OnSetPetLoadOut(slotid, petid)
    if self.db.profile.overrideSetLoadout and slotid == 1 then
        self:ResummonPet()
    end
end

function addon:HasPet()
    return C_PetJournal.GetSummonedPetID()
end

local function pick_flat(self, petlist)
    petlist = petlist or self.usable_pets
    local randomid = petlist[math.random(#petlist)]
    return randomid
end
addon.PickPetFromList = pick_flat

local function pick_weighted(self, countdown)
    countdown = (countdown or 1000)     -- upper bound on tries
    
    local randomid = self.usable_pets[math.random(#self.usable_pets)]
    local weight = self.db.profile.weights[randomid] or 1

    assert(weight > 0)
    
    if(math.random() > weight and countdown > 0) then
        -- retry
        return pick_weighted(self, countdown - 1)
    end

    return randomid
end

function addon:PickPet()
    for name, module in self:IterateModules() do
        local f = module["PickPet"]
        if f then
            local id = f(module)
            if id then 
                return id
            end
        end
    end

    if #self.usable_pets > 0 then
        if not self.db.weightedPets then
            return pick_flat(self)
        else
            return pick_weighted(self)
        end
    end
end

function addon:SummonPet()
    local petid = self:PickPet()
    if self:HasPet() == petid then
        return false
    end

    if self.db.profile.verbose then
        local _, customName, _, _, _, _, petName = C_PetJournal.GetPetInfoByPetID(petid)
        self:Printf("Summoning %s", customName or petName)
    end

    C_PetJournal.SummonPetByID(petid)
    
    self.SwitchTimer:Start()
    
    return true
end

function addon:DesummonPet(disable, verify)
    -- TODO scan old critter list to see which critters can 
    -- be desummoned using DismissCompanion("CRITTER")?
    local curid = C_PetJournal.GetSummonedPetID()
    if curid ~= nil and curid ~= 0 then
        C_PetJournal.SummonPetByID(curid)
        
        if verify then
            local handle
            local function verifyfunc()
                if C_PetJournal.GetSummonedPetID() == curid then
                    C_PetJournal.SummonPetByID(curid)
                else
                    self:CancelTimer(handle)
                end
            end
            handle = self:ScheduleRepeatingTimer(verifyfunc, 0.3)
        end
    end

    if disable then
        addon:EnableSummoning(false)
    end
end

function addon:ResummonPet(enable, nowait)
    if enable then
        self:EnableSummoning(true)
    end

    self.Ready:QueueResummon(nowait)
end

function addon:TogglePet()
    if self:HasPet() then
        self:DesummonPet(true)
    else
        self:ResummonPet(true, true)
    end
end

